from copy import deepcopy
import logging
from typing import List, Optional
from .flow import SimFlow
from ..tool import Tool

log = logging.getLogger(__name__)


class Vcs(SimFlow):
    """Synopsys VCS simulator"""

    dc_shell = Tool(executable="dc_shell-xg-t")  # type: ignore
    vlogan = Tool("vlogan")
    vhdlan = Tool("vhdlan")
    vcs = Tool("vcs")

    class Settings(SimFlow.Settings):
        simv: str = "simv"
        work_dir: Optional[str] = "work"
        timescale: Optional[str] = "1ns/1ps"
        vlogan_flags: List[str] = ["-full64", "-nc", "+warn=all"]  # TODO
        vhdlan_flags: List[str] = ["-full64", "-nc"]  # "-cycle", "-event" "+warn=all"?
        vcs_flags: List[str] = ["-full64", "-nc"]
        vcs_log_file: Optional[str] = "vcs.log"

    def run(self):
        assert isinstance(self.settings, self.Settings)
        ss = self.settings
        with open("synopsys_sim.setup", "w", encoding="utf-8") as f:
            f.writelines(
                [
                    "WORK  > default",
                    f"default : {ss.work_dir}",
                ]
            )
        vlogan_args = deepcopy(ss.vlogan_flags)
        vhdlan_args = deepcopy(ss.vhdlan_flags)
        if ss.work_dir:
            vlogan_args.extend(["-work", ss.work_dir])
            # vhdlan_args.extend(["-work", ss.work_dir]) # ???
        if ss.timescale:
            vlogan_args.append(f"-timescale={ss.timescale}")
            vhdlan_args.append(f"-timescale={ss.timescale}")
        # FIXME
        incdirs: List[str] = []
        for d in incdirs:
            vlogan_args.append(f"+incdir+{d}")
        vhdlan_args.append("-cycle")  # ???
        vhdlan_args.extend("-event")  # ???
        if self.design.language.vhdl.standard == "2008":
            vhdlan_args.extend("-vhdl08")

        # FIXME
        vlogan_args += ["-sverilog", "+v2k"]
        top = self.design.tb.top[0]
        vcs_args = [top, *ss.vcs_flags]
        vcs_args += [
            f"-j{ss.nthreads}",
            "-notice",
            "+lint=all,noVCDE,noTFIPC,noIWU,noOUDPE",
        ]

        if ss.vcs_log_file:
            vcs_args += ["-l", ss.vcs_log_file]

        log.info("analyzing Verilog files")
        self.vlogan.run(*vlogan_args)
        log.info("analyzing VHDL files")
        self.vhdlan.run(*vhdlan_args)
        log.info("Running vcs")
        self.vcs.run(*vcs_args)

        simv = Tool("./simv")  # ???
        simv_args = ["-nc", "-lca"]
        for k, v in self.design.tb.parameters.items():
            simv_args.append(f"+define+{k}={v}")
        simv_args.extend(["-l", top])
        simv.run(*simv_args)



import logging
from pathlib import Path
from typing import Any, Dict, Mapping
from box import Box
import re
import os
from ..flow import SynthFlow
from ...utils import dict_merge, toml_load, try_convert
from ...tool import Tool

log = logging.getLogger(__name__)


def get_hier(dct, dotted_path, default=None):
    splitted = dotted_path.split(".")
    merged_leaves = {}
    for i, key in enumerate(splitted):
        try:
            for k, v in dct.items():
                if not isinstance(v, Mapping):
                    merged_leaves[k] = v
            dct = dct[key]
        except KeyError:
            print(f'Key {key} not found in {".".join(splitted[:i])}!')
            return default

    for k, v in dct.items():
        # if not isinstance(v, Mapping):
        merged_leaves[k] = v
    return Box(merged_leaves)


class Dc(SynthFlow):
    dc_shell = Tool(executable="dc_shell-xg-t")  # type: ignore

    class Settings(SynthFlow.Settings):
        adk: str

    def run(self):
        assert isinstance(self.settings, self.Settings)
        ss = self.settings
        adk_id = ss.adk
        adk_root = Path.home() / "adk"
        adk_config: Dict[str, Any] = {}
        for toml_file in adk_root.glob("*.toml"):
            adk_config = dict_merge(adk_config, toml_load(toml_file), add_new_keys=True)

        adk = get_hier(adk_config, adk_id)
        assert adk is not None
        if not os.path.isabs(adk.path):
            adk.path = os.path.join(str(adk_root), adk.path)

        log.debug("adk=%s", adk)

        script_path = self.copy_from_template(
            "run.tcl",
            adk=adk,
        )
        self.dc_shell.run("-64bit", "-topographical_mode", "-f", script_path)

    def parse_reports(self) -> bool:
        reports_dir = self.reports_dir
        top_name = self.design.rtl.top
        failed = False
        self.parse_report_regex(
            reports_dir / f"{top_name}.mapped.area.rpt",
            r"Number of ports:\s*(?P<num_ports>\d+)",
            r"Number of nets:\s*(?P<num_nets>\d+)",
            r"Number of cells:\s*(?P<num_cells>\d+)",
            r"Number of combinational cells:\s*(?P<num_cells_combinational>\d+)",
            r"Number of sequential cells:\s*(?P<num_cells_sequentual>\d+)",
            r"Number of macros/black boxes:\s*(?P<num_macro_bbox>\d+)",
            r"Number of buf/inv:\s*(?P<num_buf_inv>\d+)",
            r"Number of references:\s*(?P<num_refs>\d+)",
            r"Combinational area:\s*(?P<area_combinational>\d+(?:\.\d+)?)",
            r"Buf/Inv area:\s*(?P<area_buf_inv>\d+(?:\.\d+)?)",
            r"Noncombinational area:\s*(?P<area_noncombinational>\d+(?:\.\d+)?)",
            r"Macro/Black Box area:\s*(?P<area_macro_bbox>\d+(?:\.\d+)?)",
            r"Net Interconnect area:\s*(?P<area_interconnect>\S+.*$)",
            r"Total cell area:\s*(?P<area_cell_total>\d+(?:\.\d+)?)",
            r"Total area:\s*(?P<area_macro_bbox>\w+)",
            r"Core Area:\s*(?P<area_core>\d+(?:\.\d+)?)",
            r"Aspect Ratio:\s*(?P<aspect_ratio>\d+(?:\.\d+)?)",
            r"Utilization Ratio:\s*(?P<utilization_ratio>\d+(?:\.\d+)?)",
            dotall=False,
        )

        reportfile_path = reports_dir / f"{top_name}.mapped.qor.rpt"

        def parse_kvs(kvs):
            kvs = re.split(r"\s*\n\s*", kvs)
            kvs = [re.split(r"\s*:\s*", s.strip()) for s in kvs if s.strip()]
            return {s[0].strip(): try_convert(s[1]) for s in kvs}

        path_group_re = re.compile(
            r"^\s*Timing Path Group\s+'(?P<path_group_name>\w+)'\n\s*\-+\s*\n(?P<kv>(?:^.*\n)+)",
            re.MULTILINE,
        )

        area_re = re.compile(
            r"^\s*Area\s*\n\s*\-+\s*\n(?P<kv1>(?:^.*\n)+)\s*\-+\s*\n(?P<kv2>(?:^.*\n)+)",
            re.MULTILINE,
        )
        drc_re = re.compile(
            r"^\s*Design Rules\s*\n\s*\-+\s*\n(?P<kv>(?:^.*\n)+)", re.MULTILINE
        )
        wns_re = re.compile(
            r"^\s*Design\s+WNS:\s+(?P<wns>\d+\.\d+)\s+TNS:\s+(?P<tns>\d+\.\d+)\s+Number of Violating Paths:\s*(?P<nvp>\d+)"
        )
        hold_wns_re = re.compile(
            r"^\s*Design\s+\(Hold\)\s+WNS:\s+(?P<wns>\d+\.\d+)\s+TNS:\s+(?P<tns>\d+\.\d+)\s+Number of Violating Paths:\s*(?P<nvp>\d+)"
        )

        # placeholder for ordering
        self.results["path_groups"] = None

        with open(reportfile_path) as rpt_file:
            content = rpt_file.read()
            sections = re.split(r"\n\s*\n", content)

            path_groups = {}
            for sec in sections:
                match = path_group_re.match(sec)
                if match:
                    group_name = match.group("path_group_name")
                    path_groups[group_name] = parse_kvs(match.group("kv"))

                else:
                    match = area_re.match(sec)
                    if match:
                        kv1 = parse_kvs(match.group("kv1"))
                        kv2 = parse_kvs(match.group("kv2"))
                        self.results["area"] = {**kv1, **kv2}
                    else:
                        match = drc_re.match(sec)
                        if match:
                            drc = parse_kvs(match.group("kv"))
                            self.results["drc"] = drc
                            if drc["Nets With Violations"] != 0:
                                print(
                                    f"Nets With DRC Violations: {drc['Nets With Violations']}"
                                )
                        else:
                            match = wns_re.match(sec)
                            if match:
                                self.results["wns"] = float(match.group("wns"))
                                self.results["tns"] = float(match.group("tns"))
                                self.results["num_violating_paths"] = int(
                                    match.group("nvp")
                                )
                                if (
                                    self.results["wns"] > 0
                                    or self.results["tns"] > 0
                                    or self.results["num_violating_paths"] != 0
                                ):
                                    failed = True
                            else:
                                match = hold_wns_re.match(sec)
                                if match:
                                    self.results["hold_wns"] = float(match.group("wns"))
                                    self.results["hold_tns"] = float(match.group("tns"))
                                    self.results["hold_num_violating_paths"] = int(
                                        match.group("nvp")
                                    )
                                    if (
                                        self.results["hold_wns"] > 0
                                        or self.results["hold_tns"] > 0
                                        or self.results["hold_num_violating_paths"] != 0
                                    ):
                                        failed = True
            self.results["path_groups"] = path_groups

        return not failed



import os
import random

import cocotb
from box import Box
from cocolight import DUT, DutClock, DutReset, ValidReadyTb, cocotest

NUM_TV = int(os.environ.get("NUM_TV", 2000))
DEBUG = bool(os.environ.get("DEBUG", False))


class SqrtTb(ValidReadyTb):
    def __init__(self, dut: DUT, debug: bool = DEBUG):
        super().__init__(dut, DutClock("clk"), DutReset("rst"), debug)
        self.in_bus = self.driver("radicand", data_suffix=None)
        self.out_bus = self.monitor("root", data_suffix=[None, "remainder"])

    async def verify(self, rad: int):
        stimulus = cocotb.start_soon(self.in_bus.enqueue(rad))
        out = await self.out_bus.dequeue()
        await stimulus

        assert isinstance(out, Box)
        root = int(out[None])
        remainder = int(out.remainder)

        assert rad == root**2 + remainder, f"{rad} !=  {root} ** 2 + {remainder}"
        assert rad < (root + 1) ** 2, "returned root was smaller than expected"

        self.log.debug("radicand=%d got root=%d remainder=%d", rad, root, remainder)


@cocotest()
async def test_sqrt_corners(dut: DUT, debug=False):
    tb = SqrtTb(dut, debug=debug)
    await tb.reset()
    tb.log.info("DUT: %s", str(dut))
    # get bound parameters/generics from the simulator
    G_IN_WIDTH = tb.get_int_value("G_IN_WIDTH")
    tb.log.info(f"G_IN_WIDTH:{G_IN_WIDTH}")
    # await tb.clock_edge
    # TODO make sure ranges are within G_IN_WIDTH bits
    assert G_IN_WIDTH and G_IN_WIDTH >= 6
    testcases = list(range(30)) + list(range(2**G_IN_WIDTH - 70, 2**G_IN_WIDTH))
    testcases += list(range(2**G_IN_WIDTH // 2 - 30, 2**G_IN_WIDTH // 2 + 30))
    for rad in testcases:
        await tb.verify(rad)


@cocotest
async def test_sqrt(dut: DUT, num_tests: int = NUM_TV, debug=DEBUG):
    tb = SqrtTb(dut, debug=debug)
    await tb.reset()
    # get bound parameters/generics from the simulator
    G_IN_WIDTH = tb.get_value("G_IN_WIDTH", int)
    tb.log.info("G_IN_WIDTH:%d num_tests:%d", G_IN_WIDTH, num_tests)
    # await tb.clock_edge
    testcases = [random.getrandbits(G_IN_WIDTH) for _ in range(num_tests)]
    testcases = list(set(testcases))
    for rad in testcases:
        await tb.verify(rad)



import logging
from typing import (
    Any,
    Dict,
    Union,
    Optional,
)
import pkg_resources
import os

from .utils import toml_loads, toml_load
from .dataclass import root_validator
from .flows.flow import FpgaSynthFlow, FPGA


__all__ = [
    "get_board_file_path",
    "get_board_data",
    "WithFpgaBoardSettings",
]

log = logging.getLogger(__name__)


def get_board_file_path(file: str):
    return pkg_resources.resource_filename("xeda.data.boards", file)


def get_board_data(
    board: Optional[str], custom_toml_file: Union[None, str, os.PathLike] = None
) -> Optional[Dict[str, Any]]:
    if not board:
        return None
    boards_data = {}
    if custom_toml_file:
        log.debug("Retrieving board data for %s from %s", board, custom_toml_file)
        boards_data = toml_load(custom_toml_file)
    else:
        log.info("Retrieving board data for %s", board)
        board_toml_bytes = pkg_resources.resource_string("xeda.data", "boards.toml")
        if board_toml_bytes:
            boards_data = toml_loads(board_toml_bytes.decode("utf-8"))
        else:
            log.error(
                "Unable to get resource %s.%s. Please check xeda installation.",
                "xeda.data",
                "boards.toml",
            )
    return boards_data.get(board)


class WithFpgaBoardSettings(FpgaSynthFlow.Settings):
    board: Optional[str] = None
    custom_boards_file: Optional[str] = None

    @root_validator(pre=True)
    def fpga_validate(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        board_name = values.get("board")
        fpga = values.get("fpga")
        if not fpga and board_name:
            board_data = get_board_data(board_name)
            if board_data:
                print("here4!", board_data)
                board_fpga = board_data.get("fpga")
                log.info("FPGA info for board %s: %s", board_name, str(board_fpga))
                if board_fpga:
                    values["fpga"] = FPGA(board_fpga)
        return values



# © 2020 [Kamyar Mohajerani](mailto:kamyar@ieee.org)

from typing import Optional
from ...utils import SDF
from ..flow import SimFlow
from ...tool import Tool


class Modelsim(SimFlow):
    class Settings(SimFlow.Settings):
        sdf: SDF = SDF()
        modelsimini: Optional[str] = None

    def run(self) -> None:
        assert isinstance(self.settings, self.Settings)
        vcom_options = ["-lint"]
        vlog_options = ["-lint"]
        vsim_opts = []
        tb = self.design.tb
        ss = self.settings
        # TODO are library paths supported?
        vsim_opts.extend([f"-L {lib_name}" for lib_name in ss.lib_paths])
        sdf_root = ss.sdf.root if ss.sdf.root else tb.uut
        for dt, f in ss.sdf.delay_items():
            assert sdf_root, "Neither settings.sdf.root or design.tb.uut are provided"
            vsim_opts.extend([f"-sdf{dt}", f"{sdf_root}={f}"])

        tb_generics_opts = " ".join([f"-g{k}={v}" for k, v in tb.parameters.items()])

        script_path = self.copy_from_template(
            "run.tcl",
            generics_options=tb_generics_opts,
            vcom_opts=" ".join(vcom_options),
            vlog_opts=" ".join(vlog_options),
            vsim_opts=" ".join(vsim_opts),
        )

        modelsim_opts = ["-batch", "-do", f"do {script_path}"]
        if ss.modelsimini:
            modelsim_opts.extend(["-modelsimini", ss.modelsimini])
        vsim = Tool("vsim")
        vsim.run(*modelsim_opts)



"""Interchangable dataclass abstraction"""
import logging
from abc import ABCMeta
from typing import Any, Callable, Dict, List, Optional, Tuple, Type

import attrs

# pylint: disable=no-name-in-module
from pydantic import (  # pylint: disable=no-name-in-module
    BaseConfig,  # pylint: disable=no-name-in-module
    BaseModel,  # pylint: disable=no-name-in-module
    Extra,
    Field,
    ValidationError,
    root_validator,
    validator,
)  # pylint: disable=no-name-in-module

__all__ = [
    "XedaBaseModel",
    "XedaBaseModeAllowExtra",
    "Field",
    "validator",
    "root_validator",
    "Extra",
    "asdict",
    "define",
    "ValidationError",
    "validation_errors",
]


log = logging.getLogger(__name__)


# WIP: interchangable dataclass backend
def define(maybe_cls: Optional[Type[Any]], **kwargs: Any) -> Any:
    return attrs.define(maybe_cls, **kwargs)


def field(
    default: Any = attrs.NOTHING,
    *,
    description: Optional[str] = None,
    validator_: Optional[Callable[..., None]] = None,
    converter: Optional[Callable[..., Any]] = None,
    factory: Optional[Callable[[], Any]] = None,
    on_setattr: Any = None,
    **kwargs: Any,
) -> Any:
    metadata = None
    if description is not None:
        metadata = {"description": description}
    return attrs.field(
        default=default,
        validator=validator_,
        converter=converter,
        factory=factory,
        on_setattr=on_setattr,
        metadata=metadata,
        **kwargs,
    )


def asdict(inst: Any, filter_: Optional[Callable[..., bool]] = None) -> Dict[str, Any]:
    if isinstance(inst, BaseModel):
        assert filter_ is None
        return inst.dict()
    return attrs.asdict(inst, filter=filter_)


class XedaBaseModel(BaseModel):
    class Config(BaseConfig):
        validate_assignment = True
        extra = Extra.forbid
        arbitrary_types_allowed = True
        use_enum_values = True

    # def validate(self):
    #     *_, validation_error = validate_model(self.__class__, self.__dict__)
    #     if validation_error:
    #         raise validation_error


class XedaBaseModeAllowExtra(XedaBaseModel, metaclass=ABCMeta):
    class Config(XedaBaseModel.Config):
        extra = Extra.allow


def error_type_and_context(error: Dict[str, Any]) -> str:
    ctx = error.get("ctx", {})
    return str(error.get("type")) + "".join(f"; {k}={v}" for k, v in ctx.items())


def validation_errors(errors: List[Dict[str, Any]]) -> List[Tuple[str, str, str]]:
    return [
        (
            " -> ".join(str(loc) for loc in e.get("loc", [])),
            e.get("msg", "???"),
            error_type_and_context(e),
        )
        for e in errors
    ]



from pathlib import Path
from typing import Any, Dict

from xeda.flows.flow import Cocotb
from xeda.utils import WorkingDirectory

TESTS_DIR = Path(__file__).parent.absolute()
RESOURCES_DIR = TESTS_DIR / "resources"


def test_cocotb_version():
    cocotb = Cocotb(sim_name="ghdl")  # type: ignore

    assert cocotb.version_gte(0)
    assert cocotb.version_gte(0, 0)
    assert cocotb.version_gte(0, 1)
    assert cocotb.version_gte(0, 1, 1)
    assert cocotb.version_gte(0, 1, 1, 1)
    assert cocotb.version_gte(1)
    assert cocotb.version_gte(1, 5)
    assert cocotb.version_gte(1, 6)
    assert cocotb.version_gte(1, 6, 0)
    assert cocotb.version_gte(1, 6, 1)
    assert cocotb.version_gte(1, 6, 2)
    assert not cocotb.version_gte(1, 7)
    assert not cocotb.version_gte(1, 7, 0)
    assert not cocotb.version_gte(2)
    assert not cocotb.version_gte(2, 0, 0)
    assert not cocotb.version_gte(2, 0, 0, 0)


def test_cocotb_parse_xml():
    assert (RESOURCES_DIR / "cocotb" / "results.xml").exists()
    with WorkingDirectory(RESOURCES_DIR / "cocotb"):
        cocotb = Cocotb(sim_name="dummy")  # type: ignore
        for case in cocotb.result_testcases:
            print(case)
        results: Dict[str, Any] = {}
        cocotb.add_results(results)
        if not results["success"] is False:
            raise AssertionError()


if __name__ == "__main__":
    test_cocotb_version()
    test_cocotb_parse_xml()



import json
import logging
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional

from box import Box

from ...dataclass import Field, root_validator, validator
from ...design import Design
from ...flows.ghdl import GhdlSynth
from ...tool import Docker, Tool
from ..flow import SynthFlow

log = logging.getLogger(__name__)


def append_flag(flag_list: List[str], flag: str) -> List[str]:
    if flag not in flag_list:
        flag_list.append(flag)
    return flag_list


class Yosys(SynthFlow):
    """Synthesize the design using Yosys Open SYnthesis Suite"""

    class Settings(SynthFlow.Settings):
        log_file: Optional[str] = "yosys.log"
        flatten: bool = Field(True, description="flatten design")
        abc9: bool = Field(True, description="Use abc9")
        retime: bool = Field(False, description="Enable flip-flop retiming")
        nobram: bool = Field(False, description="Do not map to block RAM cells")
        nodsp: bool = Field(False, description="Do not use DSP resources")
        nolutram: bool = Field(False, description="Do not use LUT RAM cells")
        sta: bool = Field(
            False,
            description="Run a simple static timing analysis (requires `flatten`)",
        )
        nowidelut: bool = Field(
            False,
            description="Do not use MUX resources to implement LUTs larger than native for the target",
        )
        dff: bool = Field(True, description="Run abc/abc9 with -dff option")
        widemux: int = Field(
            0,
            description="enable inference of hard multiplexer resources for muxes at or above this number of inputs"
            " (minimum value 2, recommended value >= 5 or disabled = 0)",
        )
        read_verilog_flags: List[str] = [
            "-noautowire",
        ]
        read_liberty_flags: List[str] = []
        synth_flags: List[str] = []
        abc_flags: List[str] = []
        write_vhdl_flags: List[str] = [
            # '-norename',
            # '-std08'.
            # '-noattr',
            "-renameprefix YOSYS_n",
            # "-v",
        ]
        write_verilog_flags: List[str] = [
            # '-norename',
            # '-attr2comment',
            # '-noexpr',
            # '-siminit',
            # '-extmem',
            # '-sv',
        ]
        check_assert: bool = True
        netlist_verilog: Optional[str] = "netlist.v"
        netlist_vhdl: Optional[str] = None  # "netlist.vhdl"
        netlist_json: Optional[str] = "netlist.json"
        rtl_verilog: Optional[str] = "rtl.v"
        rtl_vhdl: Optional[str] = None  # "rtl.vhdl"
        rtl_json: Optional[str] = "rtl.json"
        show_rtl: bool = False
        show_rtl_flags: List[str] = [
            "-stretch",
            "-enum",
            "-width",
        ]
        show_netlist: bool = False
        show_netlist_flags: List[str] = ["-stretch", "-enum"]
        ghdl: GhdlSynth.Settings = GhdlSynth.Settings()
        post_synth_opt: bool = Field(
            True,
            description="run additional optimization steps after synthesis if complete",
        )
        verilog_lib: List[str] = []
        splitnets: Optional[List[str]] = None  # ['-driver']
        set_attributes: Dict[str, Dict[str, Any]] = {}
        stop_after: Optional[Literal["rtl"]]
        netlistsvg: Optional[str] = Field(
            None,
            description="Generate a netlist SVG by runnning 'netlistsvg' (netlistsvg needs to be installed)",
        )

        @validator("write_verilog_flags", pre=False)
        def validate_write_verilog_flags(cls, value, values):  # type: ignore
            if values.get("debug"):
                value.append("-norename")
            return value

        @validator("verilog_lib", pre=True)
        def validate_verilog_lib(cls, value, values):  # type: ignore
            if isinstance(value, str):
                value = [value]
            value = [str(Path(v).resolve(strict=True)) for v in value]
            return value

        @validator("splitnets", pre=True)
        def validate_splitnets(cls, value, values):  # type: ignore
            if value is not None:
                if isinstance(value, str):
                    value = [value]
                value = [
                    v if not v.strip() or v.startswith("-") else f"-{v}" for v in value
                ]
            return value

        @validator("set_attributes", pre=True, always=True)
        def validate_set_attributes(cls, value, values):  # type: ignore
            if value:
                if isinstance(value, str):
                    if value.endswith(".json"):
                        attr_file = Path(value)
                        try:
                            log.info("Parsing %s as JSON file", attr_file)
                            with open(attr_file) as f:
                                value = {**json.load(f)}
                        except json.JSONDecodeError as e:
                            raise ValueError(
                                f"Decoding of JSON file {attr_file} failed: {e.args}"
                            ) from e
                        except TypeError as e:
                            raise ValueError(f"JSON TypeError: {e.args}") from e
                    else:
                        raise ValueError(
                            f"Unsupported extension for JSON file: {value}"
                        )
                for attr, attr_dict in value.items():
                    assert attr
                    assert attr_dict, "attr_dict must be a non-empty Dict[str, Any]"
                    for (path, v) in attr_dict.items():
                        assert path and v
                        if isinstance(path, list):
                            path = "/".join(path)
                        if isinstance(v, str):
                            v = f'"{v}"'
                        attr_dict[path] = v
                    value[attr] = dict(attr_dict)
            return value

        @root_validator(pre=True)
        def check_target(cls, values):  # type: ignore
            assert values.get("fpga") or values.get(
                "tech"
            ), "ERROR in flows.yosys.settings: No targets specified! Either 'fpga' or 'tech' must be specified."
            return values

    def __init__(self, flow_settings: Settings, design: Design, run_path: Path):
        super().__init__(flow_settings, design, run_path)
        self.stat_report = "utilization.rpt"
        self.timing_report = "timing.rpt"
        self.artifacts = Box(
            # non-plural names are preferred for keys, e.g. "digram" instead of "diagrams".
            diagram={},
            report=dict(utilization=self.stat_report, timing=self.timing_report),
            rtl={},
            netlist={},
        )

    def init(self) -> None:
        assert isinstance(self.settings, self.Settings)

        ss = self.settings
        if ss.fpga:
            assert ss.fpga.family or ss.fpga.vendor == "xilinx"
        if ss.rtl_json:
            self.artifacts.rtl.json = ss.rtl_json
        if ss.rtl_vhdl:
            self.artifacts.rtl.vhdl = ss.rtl_vhdl
        if ss.rtl_verilog:
            self.artifacts.rtl.verilog = ss.rtl_verilog
        if not ss.stop_after:  # FIXME
            if ss.netlist_json:
                self.artifacts.netlist.json = ss.netlist_json
            if ss.netlist_vhdl:
                self.artifacts.netlist.vhdl = ss.netlist_vhdl
            if ss.netlist_verilog:
                self.artifacts.netlist.verilog = ss.netlist_verilog

    def run(self) -> None:
        assert isinstance(self.settings, self.Settings)
        yosys = Tool(
            executable="yosys",
            docker=Docker(
                image="hdlc/impl"
            ),  # pyright: reportGeneralTypeIssues=none
        )
        ss = self.settings
        if ss.sta:
            ss.flatten = True

        if ss.flatten:
            append_flag(ss.synth_flags, "-flatten")

        # add FPGA-specific synth_xx flags
        if ss.fpga:
            if ss.abc9:  # ABC9 is for only FPGAs?
                append_flag(ss.synth_flags, "-abc9")
            if ss.retime:
                append_flag(ss.synth_flags, "-retime")
            if ss.dff:
                append_flag(ss.synth_flags, "-dff")
            if ss.nobram:
                append_flag(ss.synth_flags, "-nobram")
            if ss.nolutram:
                append_flag(ss.synth_flags, "-nolutram")
            if ss.nodsp:
                append_flag(ss.synth_flags, "-nodsp")
            if ss.nowidelut:
                append_flag(ss.synth_flags, "-nowidelut")
            if ss.widemux:
                append_flag(ss.synth_flags, f"-widemux {ss.widemux}")
            if ss.fpga.vendor == "xilinx":
                ss.write_vhdl_flags.append("-unisim")
        else:
            if ss.dff:
                append_flag(ss.abc_flags, "-dff")
            if ss.tech:
                if ss.tech.gates:
                    append_flag(ss.abc_flags, f"-g {ss.tech.gates}")
                elif ss.tech.liberty:
                    liberty_path = Path(ss.tech.liberty)
                    liberty_path = liberty_path.resolve().absolute()
                    assert (
                        liberty_path.exists()
                    ), f"Specified tech.liberty={liberty_path} does not exist!"
                    ss.tech.liberty = str(liberty_path)
                    append_flag(ss.abc_flags, f"-liberty {ss.tech.liberty}")
                elif ss.tech.lut:
                    append_flag(ss.abc_flags, f"-lut {ss.tech.lut}")

        script_path = self.copy_from_template(
            "yosys.tcl",
            ghdl_args=GhdlSynth.synth_args(ss.ghdl, self.design),
        )
        log.info(
            "Yosys script: %s", self.run_path.relative_to(Path.cwd()) / script_path
        )
        # args = ['-s', script_path]
        args = ["-c", script_path]
        if ss.log_file:
            args.extend(["-L", ss.log_file])
        if not ss.verbose:  # reduce noise unless verbose
            args.extend(["-T", "-Q", "-q"])
        self.results["_tool"] = yosys.info  # TODO where should this go?
        log.info("Logging yosys output to %s", ss.log_file)
        yosys.run(*args)
        skin_file = None
        elk_layout = None
        if ss.netlistsvg:
            netlistsvg = Tool("netlistsvg")
            rtl_json = self.artifacts.rtl.json
            if rtl_json:
                svg_file = "rtl.svg"
                self.artifacts.diagram.netlistsvg_rtl = svg_file
                args = [rtl_json, "-o", svg_file]
                if skin_file:
                    args.extend(["--skin", skin_file])
                if elk_layout:
                    args.extend(["--layout", elk_layout])
                netlistsvg.run(*args)
            if ss.netlist_json:
                svg_file = "netlist.svg"
                self.artifacts.diagram["netlistsvg"] = svg_file
                args = [self.artifacts.netlist.json, "-o", svg_file]
                if skin_file:
                    args.extend(["--skin", skin_file])
                if elk_layout:
                    args.extend(["--layout", elk_layout])
                netlistsvg.run(*args)

    def parse_reports(self) -> bool:
        assert isinstance(self.settings, self.Settings)
        if self.settings.fpga:
            if self.settings.fpga.vendor == "xilinx":
                self.parse_report_regex(
                    self.artifacts.report.utilization,
                    r"=== design hierarchy ===",
                    r"FDRE\s*(?P<_FDRE>\d+)",
                    r"FDSE\s*(?P<_FDSE>\d+)",
                    r"number of LCs:\s*(?P<Estimated_LCs>\d+)",
                    sequential=True,
                    required=False,
                )
                self.results["FFs"] = int(self.results.get("_FDRE", 0)) + int(
                    self.results.get("_FDSE", 0)
                )
            if self.settings.fpga.family == "ecp5":
                self.parse_report_regex(
                    self.artifacts.report.utilization,
                    r"TRELLIS_FF\s+(?P<FFs>\d+)",
                    r"LUT4\s+(?P<LUT4>\d+)",
                )
        return True



# © 2020 [Kamyar Mohajerani](mailto:kamyar@ieee.org)
"""Intel Quartus flows"""
import csv
import logging
from pathlib import Path
from typing import Any, Callable, Dict, Literal, Optional, Set

from xeda.utils import try_convert

from ...dataclass import Field
from ...tool import Docker, Tool
from ...types import PathLike
from ..flow import FpgaSynthFlow

log = logging.getLogger(__name__)


def identity(x: Any) -> Any:
    return x


def parse_csv(
    path: PathLike,
    id_field: Optional[str],
    field_parser: Callable[[str], Any] = identity,
    id_parser: Callable[[str], Any] = identity,
    interesting_fields: Optional[Set[str]] = None,
) -> Dict[str, Any]:
    """Parse TCL-generated CSV file"""
    data: Dict[str, Any] = {}
    with open(path, newline="") as csvfile:
        if id_field:
            # with header, and some rows of data indexed by id_field
            reader = csv.DictReader(csvfile)
            for row in reader:
                if interesting_fields is None:
                    interesting_fields = set(row)
                row_id = row.get(id_field)
                if row_id:
                    data[id_parser(row_id)] = {
                        k: field_parser(row[k]) for k in interesting_fields if k in row
                    }
                else:
                    log.critical("ID field %s not found in row (%s)", id_field, row)
        else:
            # no header, key/value pairs on each line
            for lrow in csv.reader(csvfile):
                if len(lrow) >= 2:
                    data[id_parser(lrow[0])] = field_parser(",".join(lrow[1:]))
                else:
                    log.critical("Line is not in key,value format: %s", lrow)
    return data


def try_num(s: str):
    s = s.strip()
    try:
        return int(s)
    except ValueError:
        try:
            return float(s)
        except ValueError:
            return s


def try_float(s: str):
    s = s.strip()
    try:
        return float(s)
    except ValueError:
        return s


class Quartus(FpgaSynthFlow):
    """FPGA synthesis using Intel Quartus"""

    quartus_sh = Tool(
        executable="quartus_sh",
        docker=Docker(
            command=["quartus_wrapper", "quartus_sh"],
            image="chriz2600/quartus-lite",
            tag="20.1.0",
            platform="linux/amd64",
        ),  # type: ignore
    )

    class Settings(FpgaSynthFlow.Settings):
        # part number (fpga.part) formats are quite complicated.
        # See: https://www.intel.com/content/dam/www/central-libraries/us/en/documents/product-catalog.pdf
        seed: Optional[int] = Field(None, description="Seed")
        optimization_mode: Literal[
            "BALANCED",
            "HIGH PERFORMANCE EFFORT",
            "AGGRESSIVE PERFORMANCE",
            "High Performance with Maximum Placement Effort",
            "Superior Performance",
            "Superior Performance with Maximum Placement Effort",
            "Aggressive Area",
            "High Placement Routability Effort",
            "High Packing Routability Effort",
            "Optimize Netlist for Routability",
            "High Power Effort",
        ] = Field(
            "HIGH PERFORMANCE EFFORT",
            description="""
            see https://www.intel.com/content/www/us/en/programmable/documentation/zpr1513988353912.html
            https://www.intel.com/content/www/us/en/programmable/quartushelp/current/index.htm
        """,
        )
        remove_redundant_logic: bool = True
        auto_resource_sharing: bool = True
        retiming: bool = True
        register_duplication: bool = True
        packed_registers: bool = False
        gated_clock_conversion: bool = False
        dsp_recognition: bool = True
        ram_recognition: bool = True
        rom_recognition: bool = True
        synthesis_effort: Literal["auto", "fast"] = "auto"
        fitter_effort: Literal["STANDARD FIT", "AUTO FIT", "FAST_FIT"] = "AUTO FIT"
        optimization_technique: Literal["AREA", "SPEED", "BALANCED"] = "SPEED"
        placement_effort_multiplier: float = Field(
            2.0,
            description="""
        A logic option that controls how much time the Fitter spends in placement.
        The default value is 1.0 and legal values must be greater than 0 and can be non-integer values.
        Values between 0 and 1 can reduce fitting time, but also can reduce placement quality and design performance.
        Values greater than 1 increase placement time and placement quality, but may reduce routing time for designs with routing congestion.
        For example, a value of 4 increases fitting time by approximately 2 to 4 times, but may improve quality.""",
        )
        router_timing_optimization_level: Literal[
            "Normal", "Maximum", "Minimum"
        ] = "Maximum"
        final_placement_optimization: Literal[
            "ALWAYS", "AUTOMATICALLY", "NEVER"
        ] = "ALWAYS"

    def init(self) -> None:
        self.reports = {
            "summary": self.reports_dir / "Flow_Summary.csv",
            "utilization": self.reports_dir
            / "Fitter"
            / "Resource_Section"
            / "Fitter_Resource_Utilization_by_Entity.csv",
            "timing_dir": self.reports_dir / "Timing_Analyzer",
            "timing.multicorner_summary": self.reports_dir
            / "Timing_Analyzer"
            / "Multicorner_Timing_Analysis_Summary.csv",
        }
        # FIXME only a proof of concept. Tool instantiation/invocation needs to change!
        if self.quartus_sh.docker:
            self.quartus_sh.docker.enabled = self.settings.dockerized
            # FIXME
            for src in self.design.rtl.sources:
                p = str(src.file.parent.resolve())
                self.quartus_sh.docker.mounts[p] = p
            if self.settings.dockerized and self.quartus_sh.docker.nproc:
                self.settings.ncpus = min(
                    self.settings.ncpus, self.quartus_sh.docker.nproc
                )
                self.settings.nthreads = min(
                    self.settings.nthreads, self.quartus_sh.docker.nproc
                )

    def create_project(self, **kwargs: Any) -> None:
        assert isinstance(self.settings, self.Settings)
        ss = self.settings

        project_settings = {
            "SEED": ss.seed,
            "OPTIMIZATION_MODE": ss.optimization_mode,
            "REMOVE_REDUNDANT_LOGIC_CELLS": ss.remove_redundant_logic,
            "AUTO_RESOURCE_SHARING": ss.auto_resource_sharing,
            "ALLOW_REGISTER_RETIMING": ss.retiming,
            "SYNTH_GATED_CLOCK_CONVERSION": ss.gated_clock_conversion,
            # faster:
            "FITTER_EFFORT": ss.fitter_effort,
            # AREA, SPEED, BALANCED
            "STRATIX_OPTIMIZATION_TECHNIQUE": ss.optimization_technique,
            "CYCLONE_OPTIMIZATION_TECHNIQUE": ss.optimization_technique,
            # see https://www.intel.com/content/www/us/en/programmable/documentation/rbb1513988527943.html
            "PLACEMENT_EFFORT_MULTIPLIER": ss.placement_effort_multiplier,
            "ROUTER_TIMING_OPTIMIZATION_LEVEL": ss.router_timing_optimization_level,
            "FINAL_PLACEMENT_OPTIMIZATION": ss.final_placement_optimization,
            # "PHYSICAL_SYNTHESIS_COMBO_LOGIC_FOR_AREA": True,
            # ?
            # "ADV_NETLIST_OPT_SYNTH_GATE_RETIME": True,
            # ?
            # "ADV_NETLIST_OPT_SYNTH_WYSIWYG_REMAP": True,
            "AUTO_PACKED_REGISTERS_STRATIX": ss.packed_registers,
            "AUTO_PACKED_REGISTERS_CYCLONE": ss.packed_registers,
            "PHYSICAL_SYNTHESIS_COMBO_LOGIC": True,
            "PHYSICAL_SYNTHESIS_REGISTER_DUPLICATION": ss.register_duplication,
            "PHYSICAL_SYNTHESIS_REGISTER_RETIMING": ss.retiming,
            # "PHYSICAL_SYNTHESIS_EFFORT": "EXTRA",
            # NORMAL, OFF, EXTRA_EFFORT
            # "OPTIMIZE_POWER_DURING_SYNTHESIS": "NORMAL",
            # SYNTH_CRITICAL_CLOCK: ON, OFF : Speed Optimization Technique for Clock Domains}
            "AUTO_DSP_RECOGNITION": ss.dsp_recognition,
            "AUTO_RAM_RECOGNITION": ss.ram_recognition,
            "AUTO_ROM_RECOGNITION": ss.rom_recognition,
            "FLOW_ENABLE_POWER_ANALYZER": True,
        }

        clock_sdc_path = self.copy_from_template("clock.sdc")
        script_path = self.copy_from_template(
            "create_project.tcl",
            sdc_files=[clock_sdc_path],
            project_settings=project_settings,
            **kwargs,
        )
        self.quartus_sh.run("-t", script_path)

        # self.run_process('quartus_sh',
        #                  ['--dse', '-project', self.settings.design['name'], '-nogui', '-concurrent-compiles', '8', '-exploration-space',
        #                   "Extra Effort Space", '-optimization-goal', "Optimize for Speed", '-report-all-resource-usage', '-ignore-failed-base'],
        #                  stdout_logfile='dse_stdout.log'
        #                  )

    # def __init__(self, settings, args, logger):
    #     # def supported_quartus_generic(k, v, sim):
    #     #     if sim:
    #     #         return True
    #     #     if isinstance(v, int):
    #     #         return True
    #     #     if isinstance(v, bool):
    #     #         return True
    #     #     v = str(v)
    #     #     return (v.isnumeric() or (v.strip().lower() in {'true', 'false'}))

    #     # def quartus_gen_convert(k, x, sim):
    #     #     if sim:
    #     #         if isinstance(x, dict) and "file" in x:
    #     #             p = x["file"]
    #     #             assert isinstance(p, str), "value of `file` should be a relative or absolute path string"
    #     #             x = self.conv_to_relative_path(p.strip())
    #     #             self.logger.info(f'Converting generic `{k}` marked as `file`: {p} -> {x}')
    #     #     xl = str(x).strip().lower()
    #     #     if xl == 'false':
    #     #         return "1\\'b0"
    #     #     if xl == 'true':
    #     #         return "1\\'b1"
    #     #     return x

    #     # def quartus_generics(kvdict, sim):
    #     #     return ' '.join([f"-generic {k}={quartus_gen_convert(k, v, sim)}" for k, v in kvdict.items() if supported_quartus_generic(k, v, sim)])

    #     super().__init__(settings, args, logger)

    #     # self.settings.flow['generics_options'] = quartus_generics(self.settings.design["generics"], sim=False)
    #     # self.settings.flow['tb_generics_options'] = quartus_generics(self.settings.design["tb_generics"], sim=True)

    # self.run_process('quartus_eda', [prj_name, '--simulation', '--functional', '--tool=modelsim_oem', '--format=verilog'],
    #                         stdout_logfile='eda_1_stdout.log'
    #                         )

    def run(self) -> None:
        self.create_project()
        script_path = self.copy_from_template(
            "compile.tcl", reports_dir=self.reports_dir
        )
        self.quartus_sh.run("-t", script_path)

    def parse_reports(self) -> bool:
        assert isinstance(self.settings, self.Settings)
        failed = False
        reports = self.reports
        resources = parse_csv(reports["summary"], id_field=None)
        self.results.update(resources)
        utilization_report = Path(reports["utilization"])
        assert utilization_report.exists()
        resources = parse_csv(
            utilization_report,
            id_field="Compilation Hierarchy Node",
            field_parser=lambda s: try_num(s.split()[0]),
            id_parser=lambda s: s.strip().lstrip("|"),
            # interesting_fields={
            #     "Logic Cells",
            #     "LUT-Only LCs",
            #     "Register-Only LCs",
            #     "LUT/Register LCs",
            #     "Dedicated Logic Registers",
            #     "ALMs needed [=A-B+C]",
            #     "Combinational ALUTs",
            #     "ALMs used for memory",
            #     "Memory Bits",
            #     "M10Ks",
            #     "M9Ks",
            #     "DSP Elements",
            #     "DSP Blocks",
            #     "DSP 9x9",
            #     "DSP 18x18"
            #     "Block Memory Bits",
            #     "Pins",
            #     "I/O Registers",
            # },
        )

        top_resources = resources.get(self.design.rtl.top)
        if top_resources:
            top_resources.setdefault(0)
            lut_only = top_resources.get("LUT-Only LCs", 0)
            lut_reg = top_resources.get("LUT/Register LCs", 0)
            top_resources["lut"] = lut_only + lut_reg
            reg_only = top_resources.get("Register-Only LCs", 0)
            top_resources["ff"] = reg_only + lut_reg
            self.results.update(top_resources)

        # TODO reference for why this timing report is chosen

        mc_report = reports["timing.multicorner_summary"]

        slacks = parse_csv(
            mc_report,
            id_field="Clock",
            field_parser=try_float,
            id_parser=lambda s: s.strip(),
            interesting_fields={"Setup", "Hold"},
        )
        print("slacks:", slacks)
        worst_slacks = slacks.get("Worst-case Slack")
        print("worst_slacks:", worst_slacks)
        if worst_slacks:
            wns = worst_slacks.get("Setup")
            print("wns:", wns)
            whs = worst_slacks.get("Hold")
            if wns is not None:
                wns = try_convert(wns)
                self.results["wns"] = wns
                if isinstance(wns, (int, float)):
                    failed |= wns < 0
            if whs is not None:
                whs = try_convert(whs)
                self.results["whs"] = whs
                if isinstance(whs, (int, float)):
                    failed |= whs < 0

        timing_reports_folder = Path(reports["timing_dir"])
        max_fmax = 0.0
        for fmax_report in timing_reports_folder.glob(
            "Slow_*_Model/Slow_*_Model_Fmax_Summary.csv"
        ):
            log.info("Parsing timing report: %s", fmax_report)
            fmax = parse_csv(
                fmax_report,
                id_field="Clock Name",
                field_parser=lambda s: s.strip().split(),
                id_parser=lambda s: s.strip(),
                interesting_fields={"Fmax"},
            )
            conditions = (
                fmax_report.parent.name.lstrip("Slow_").rstrip("_Model").split("_")
            )
            for clock in self.settings.clocks.keys():
                flst = fmax.get(clock, {}).get("Fmax")
                assert flst and len(flst) == 2
                fmhz = float(flst[0])
                if flst[1] == "GHz":
                    fmhz *= 1000.0
                else:
                    assert flst[1] == "MHz"
                if fmhz > max_fmax:
                    max_fmax = fmhz
                self.results[f'{clock} Fmax@{":".join(conditions)} (MHz)'] = fmhz
        if max_fmax:
            self.results["Fmax"] = max_fmax

        return not failed


# class QuartusDse(QuartusSynth, DseFlow):
#     def run(self):
#         self.create_project()
#         # 'explore': Exploration flow to use, if not specified in --config
#         #   configuration file. Valid flows: timing_aggressive,
#         #   all_optimization_modes, timing_high_effort, seed,
#         #   area_aggressive, power_high_effort, power_aggressive
#         # 'compile_flow':  'full_compile', 'fit_sta' and 'fit_sta_asm'.
#         # 'timeout': Limit the amount of time a compute node is allowed to run. Format: hh:mm:ss
#         if 'dse' not in self.settings.flow:
#             self.fatal('`flows.quartus.dse` settings are missing!')

#         dse = self.settings.flow['dse']
#         if 'nproc' not in dse or not dse['nproc']:
#             dse['nproc'] = self.nthreads

#         script_path = self.copy_from_template(f'settings.dse',
#                                               dse=dse
#                                               )
#         self.run_process('quartus_dse',
#                          ['--use-dse-file', script_path, self.settings.design['name']],
#                          stdout_logfile='dse_stdout.log',
#                          initial_step="Running Quartus DSE",
#                          )

# DES:

# Available exploration spaces for this family are:
# "Seed Sweep"
# "Extra Effort Space"
# "Extra Effort Space for Quartus Prime Integrated Synthesis Projects"
# "Area Optimization Space"
# "Signature: Placement Effort Multiplier"
# "Custom Space"

# Valid optimization-goal options are:
# "Optimize for Speed"
# "Optimize for Area"
# "Optimize for Power"
# "Optimize for Negative Slack and Failing Paths"
# "Optimize for Average Period"
# "Optimize for Quality of Fit"



import logging
import os
from pathlib import Path
from typing import Any, Dict, List, Optional

from ...dataclass import Field, XedaBaseModel
from ..flow import FpgaSynthFlow
from ..vivado import Vivado

log = logging.getLogger(__name__)


StepsValType = Optional[Dict[str, Any]]


class RunOptions(XedaBaseModel):
    strategy: Optional[str] = None
    steps: Dict[str, StepsValType] = {}


class VivadoSynth(Vivado, FpgaSynthFlow):
    """Synthesize with Xilinx Vivado using a project-based flow"""

    class Settings(Vivado.Settings, FpgaSynthFlow.Settings):
        """Settings for Vivado synthesis in project mode"""

        # FIXME implement and verify all
        checkpoints_dir: str = "checkpoints"
        fail_critical_warning = Field(
            False,
            description="flow fails if any Critical Warnings are reported by Vivado",
        )
        fail_timing = Field(True, description="flow fails if timing is not met")
        optimize_power = False
        optimize_power_postplace = False
        blacklisted_resources: List[str] = Field(  # TODO: remove
            # ["latch"],
            [],
            description="list of FPGA resources which are not allowed to be inferred or exist in the results. Valid values: latch, dsp, bram",
        )
        input_delay: Optional[float] = None
        output_delay: Optional[float] = None
        out_of_context: bool = False
        write_checkpoint: bool = False
        write_netlist: bool = True
        write_bitstream: bool = False
        qor_suggestions: bool = False
        synth: RunOptions = RunOptions(
            strategy="Flow_PerfOptimized_high",
            steps={
                "SYNTH_DESIGN": {},
                "OPT_DESIGN": {},
                "POWER_OPT_DESIGN": {},
            },
        )
        impl: RunOptions = RunOptions(
            strategy="Performance_ExploreWithRemap",
            steps={
                "PLACE_DESIGN": {},
                "POST_PLACE_POWER_OPT_DESIGN": {},
                "PHYS_OPT_DESIGN": {},
                "ROUTE_DESIGN": {},
                "WRITE_BITSTREAM": {},
            },
        )

    def run(self):
        assert isinstance(self.settings, self.Settings)
        settings = self.settings
        for o in [
            "timesim.min.sdf",
            "timesim.max.sdf",
            "timesim.v",
            "funcsim.vhdl",
            "xdc",
        ]:
            self.artifacts[o] = os.path.join(settings.outputs_dir, o)

        settings.synth.steps = {
            **{
                "SYNTH_DESIGN": {},
                "OPT_DESIGN": {},
                "POWER_OPT_DESIGN": {},
            },
            **settings.synth.steps,
        }
        settings.impl.steps = {
            **{
                "PLACE_DESIGN": {},
                "POST_PLACE_POWER_OPT_DESIGN": {},
                "PHYS_OPT_DESIGN": {},
                "ROUTE_DESIGN": {},
                "WRITE_BITSTREAM": {},
            },
            **settings.impl.steps,
        }

        if not self.design.rtl.clock_port:
            log.critical(
                "No clocks specified for top RTL design. Continuing with synthesis anyways."
            )
        else:
            assert (
                settings.clock_period
            ), "`clock_period` must be specified and be positive value"
            freq = 1000 / settings.clock_period
            log.info(
                "clock.port=%s clock.frequency=%.3f MHz",
                self.design.rtl.clock_port,
                freq,
            )
        clock_xdc_path = self.copy_from_template("clock.xdc")

        if settings.blacklisted_resources:
            log.info("blacklisted_resources: %s", self.settings.blacklisted_resources)

        if settings.synth.steps["SYNTH_DESIGN"] is None:
            settings.synth.steps["SYNTH_DESIGN"] = {}
        assert settings.synth.steps["SYNTH_DESIGN"] is not None
        if "bram_tile" in settings.blacklisted_resources:
            # FIXME also add -max_uram 0 for ultrascale+
            settings.synth.steps["SYNTH_DESIGN"]["MAX_BRAM"] = 0
        if "dsp" in settings.blacklisted_resources:
            settings.synth.steps["SYNTH_DESIGN"]["MAX_DSP"] = 0

        reports_tcl = self.copy_from_template("vivado_report_helper.tcl")
        script_path = self.copy_from_template(
            "vivado_synth.tcl", xdc_files=[clock_xdc_path], reports_tcl=reports_tcl
        )
        self.vivado.run("-source", script_path)

    def parse_timing_report(self, reports_dir) -> bool:
        failed = False
        assert isinstance(self.settings, self.Settings)
        if self.design.rtl.clock_port and self.settings.clock_period:
            failed |= not self.parse_report_regex(
                reports_dir / "timing_summary.rpt",
                ##
                r"Timing\s+Summary[\s\|\-]+WNS\(ns\)\s+TNS\(ns\)\s+"
                r"TNS Failing Endpoints\s+TNS Total Endpoints\s+WHS\(ns\)\s+THS\(ns\)\s+"
                r"THS Failing Endpoints\s+THS Total Endpoints\s+WPWS\(ns\)\s+TPWS\(ns\)\s+"
                r"TPWS Failing Endpoints\s+TPWS Total Endpoints\s*"
                r"\s*(?:\-+\s+)+"
                r"(?P<wns>\-?\d+(?:\.\d+)?)\s+(?P<_tns>\-?\d+(?:\.\d+)?)\s+(?P<_failing_endpoints>\d+)\s+(?P<_tns_total_endpoints>\d+)\s+"
                r"(?P<whs>\-?\d+(?:\.\d+)?)\s+(?P<_ths>\-?\d+(?:\.\d+)?)\s+(?P<_ths_failing_endpoints>\d+)\s+(?P<_ths_total_endpoints>\d+)\s+",
                ##
                r"Clock Summary[\s\|\-]+^\s*Clock\s+.*$[^\w]+(\w*)\s+(\{.*\})\s+(?P<clock_period>\d+(?:\.\d+)?)\s+(?P<clock_frequency>\d+(?:\.\d+)?)",
                required=False,
            )
            wns = self.results.get("wns")
            if wns:
                if not isinstance(wns, (float, int)):
                    log.critical("Parse value for `WNS` is %s (%s)", wns, type(wns))
                else:
                    if wns < 0:
                        failed = True
                    # see https://support.xilinx.com/s/article/57304?language=en_US
                    # Fmax in Megahertz
                    # Here Fmax refers to: "The maximum frequency a design can run on Hardware in a given implementation
                    # Fmax = 1/(T-WNS), with WNS positive or negative, where T is the target clock period."
                    if "clock_period" in self.results:
                        clock_period = self.results["clock_period"]
                        assert isinstance(
                            clock_period, (float, int)
                        ), f"clock_period: {clock_period} is not a number"
                        self.results["Fmax"] = 1000.0 / (clock_period - wns)

        return not failed

    def parse_reports(self) -> bool:
        report_stage = "route_design"
        reports_dir = Path("reports") / report_stage

        failed = not self.parse_timing_report(reports_dir)

        report_file = reports_dir / "utilization.xml"
        utilization = self.parse_xml_report(report_file)
        # ordered dict
        fields = [
            ("slice", ["Slice Logic Distribution", "Slice"]),
            ("slice", ["CLB Logic Distribution", "CLB"]),  # Ultrascale+
            ("lut", ["Slice Logic", "Slice LUTs"]),
            ("lut", ["Slice Logic", "Slice LUTs*"]),
            ("lut", ["CLB Logic", "CLB LUTs"]),
            ("lut", ["CLB Logic", "CLB LUTs*"]),
            ("lut_logic", ["Slice Logic", "LUT as Logic"]),
            ("lut_logic", ["CLB Logic", "LUT as Logic"]),
            ("lut_mem", ["Slice Logic", "LUT as Memory"]),
            ("lut_mem", ["CLB Logic", "LUT as Memory"]),
            ("ff", ["Slice Logic", "Register as Flip Flop"]),
            ("ff", ["CLB Logic", "CLB Registers"]),
            ("latch", ["Slice Logic", "Register as Latch"]),
            ("latch", ["CLB Logic", "Register as Latch"]),
            ("bram_tile", ["Memory", "Block RAM Tile"]),
            ("bram_tile", ["BLOCKRAM", "Block RAM Tile"]),
            ("bram_RAMB36", ["Memory", "RAMB36/FIFO*"]),
            ("bram_RAMB36", ["BLOCKRAM", "RAMB36/FIFO*"]),
            ("bram_RAMB18", ["Memory", "RAMB18"]),
            ("bram_RAMB18", ["BLOCKRAM", "RAMB18"]),
            ("dsp", ["DSP", "DSPs"]),
            ("dsp", ["ARITHMETIC", "DSPs"]),
        ]
        if utilization is not None:
            for k, path in fields:
                if self.results.get(k) is None:
                    path.append("Used")
                    try:
                        self.results[k] = self.get_from_path(utilization, path)
                    except KeyError as e:
                        log.debug(
                            "determining %s: property %s (in %s) was not found in the utilization report %s.",
                            k,
                            e.args[0] if len(e.args) > 0 else "?",
                            path,
                            report_file,
                        )

        self.results["_utilization"] = utilization

        assert isinstance(self.settings, self.Settings)

        if not failed:
            for res in self.settings.blacklisted_resources:
                res_util = self.results.get(res)
                if res_util is not None:
                    try:
                        res_util = int(res_util)
                        if res_util > 0:
                            log.critical(
                                "%s utilization report lists %s use(s) of blacklisted resource: %s",
                                report_stage,
                                res_util,
                                res,
                            )
                            failed = True
                    except ValueError:
                        log.warning(
                            "Unknown utilization value: %s for blacklisted resource %s. Assuming results are not violating the blacklist criteria.",
                            res_util,
                            res,
                        )

            # TODO better fail analysis for vivado
            if "wns" in self.results:
                failed |= self.results["wns"] < 0
            if "whs" in self.results:
                failed |= self.results["whs"] < 0
            if "_failing_endpoints" in self.results:
                failed |= self.results["_failing_endpoints"] != 0

        return not failed



import logging

from xeda.flows.flow import FlowFatalError

from ...design import DesignSource, RtlSettings
from ...utils import SDF
from .vivado_sim import VivadoSim
from .vivado_synth import VivadoSynth

log = logging.getLogger(__name__)


class VivadoPostsynthSim(VivadoSim):
    """
    Synthesizes & implements the design, then runs post-synthesis/post-implementation simulation on the generated netlist.
    The netlist can be optionally annotated with generated timing information (SDF).
    Depends on VivadoSynth
    """

    class Settings(VivadoSim.Settings):
        synth: VivadoSynth.Settings
        timing_sim: bool = False
        enforce_io_delay: bool = True

    def init(self) -> None:
        ss = self.settings
        assert isinstance(ss, self.Settings)
        if ss.enforce_io_delay:
            # Force 0 I/O delay if no I/O delay is given
            # This seems to be required to meet timing
            if ss.synth.input_delay is None:
                ss.synth.input_delay = 0.0
            if ss.synth.output_delay is None:
                ss.synth.output_delay = 0.0
        self.add_dependency(VivadoSynth, ss.synth)

    def run(self) -> None:
        synth_flow = self.completed_dependencies[0]
        assert isinstance(synth_flow, VivadoSynth)
        ss = self.settings
        assert isinstance(ss, self.Settings)

        artifacts_path = (
            synth_flow.run_path / synth_flow.settings.outputs_dir / "route_design"
        )
        synth_netlist_path = artifacts_path / "timesim.v"
        if not synth_netlist_path.exists():
            raise FlowFatalError(f"Netlist {synth_netlist_path} does not exist!")
        postsynth_sources = [DesignSource(synth_netlist_path)]
        log.info("Setting post-synthesis sources to: %s", postsynth_sources)
        # also removing top-level generics and everything else
        self.design.rtl = RtlSettings(
            top=self.design.rtl.top, sources=postsynth_sources
        )
        assert self.design.tb and self.design.tb.top
        self.design.tb.top = (self.design.tb.top[0], "glbl")

        if "simprims_ver" not in ss.lib_paths:
            ss.lib_paths.append(("simprims_ver", None))

        if ss.timing_sim:
            if not ss.sdf.delay_items():
                ss.sdf = SDF(max=str(artifacts_path / "timesim.max.sdf"))
            if not ss.sdf.root:
                ss.sdf.root = self.design.tb.uut
            log.info("Timing simulation using SDF %s", ss.sdf)

        ss.elab_flags.extend(
            [
                "-maxdelay",
                "-transport_int_delays",
                "-pulse_r 0",
                "-pulse_int_r 0",
                "-pulse_e 0",
                "-pulse_int_e 0",
            ]
        )
        # run VivadoSim
        super().run()



"""EDA flow abstraction"""
from __future__ import annotations

import inspect
import logging
import multiprocessing
import os
import re
from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar, Union

import jinja2
import psutil
from box import Box
from jinja2 import ChoiceLoader, PackageLoader, StrictUndefined

from ..dataclass import (
    Field,
    ValidationError,
    XedaBaseModel,
    root_validator,
    validation_errors,
    validator,
)
from ..design import Design
from ..tool import Tool
from ..utils import camelcase_to_snakecase, try_convert, unique
from .cocotb import Cocotb, CocotbSettings

log = logging.getLogger(__name__)


__all__ = [
    "Flow",
    "FlowSettingsError",
    "FlowFatalError",
    "FPGA",
    "SimFlow",
    "SynthFlow",
    "Tool",
]


def regex_match(
    string: str, pattern: str, ignorecase: bool = False
) -> Optional[re.Match]:
    if not isinstance(string, str):
        return None
    return re.match(pattern, string, flags=re.I if ignorecase else 0)


def removesuffix(s: str, suffix: str) -> str:
    """similar to str.removesuffix in Python 3.9+"""
    return s[: -len(suffix)] if suffix and s.endswith(suffix) else s


def removeprefix(s: str, suffix: str) -> str:
    """similar to str.removeprefix in Python 3.9+"""
    return s[len(suffix) :] if suffix and s.startswith(suffix) else s


registered_flows: Dict[str, Tuple[str, Type["Flow"]]] = {}


DictStrPath = Dict[str, Union[str, os.PathLike]]


T = TypeVar("T", bound="Flow")


class Flow(metaclass=ABCMeta):
    """A flow may run one or more tools and is associated with a single set of settings and a single design.
    All tool executables should be available on the installed system or on the same docker image."""

    name: str  # set automatically

    class Settings(XedaBaseModel):
        """Settings that can affect flow's behavior"""

        quiet: bool = Field(False, hidden_from_schema=True)
        verbose: int = Field(0, hidden_from_schema=True)
        # debug: DebugLevel = Field(DebugLevel.NONE.value, hidden_from_schema=True)
        debug: bool = Field(False, hidden_from_schema=True)
        timeout_seconds: int = Field(3600 * 2, hidden_from_schema=True)
        nthreads: int = Field(
            default_factory=multiprocessing.cpu_count,
            description="max number of threads",
        )
        ncpus: int = Field(
            psutil.cpu_count(logical=False),
            description="Number of physical CPUs to use.",
        )
        no_console: bool = Field(False, hidden_from_schema=True)
        reports_dir: str = Field("reports", hidden_from_schema=True)
        outputs_dir: str = Field("outputs", hidden_from_schema=True)
        clean: bool = False  # TODO remove!
        lib_paths: List[
            Union[
                Tuple[
                    str,  # library name/identifier
                    Union[None, str, Path],  # optional library path
                ],
                Tuple[None, Union[str, Path]],  # or just the path
                # both name and path can't be None
            ]
        ] = Field(
            [],
            description="Additional libraries specified as a list of (name, path) tuples. Either name or path can be none. A single string or a list of string is converted to a mapping of library names without paths",
        )
        dockerized: bool = Field(False, description="Run tools from docker")

        @validator("lib_paths", pre=True)
        def lib_paths_validator(cls, value):  # pylint: disable=no-self-argument
            if isinstance(value, str):
                value = [(value, None)]
            elif isinstance(value, (list, tuple)):
                value = [(x, None) if isinstance(x, str) else x for x in value]
            return value

        def __init__(self, **data: Any) -> None:
            try:
                log.debug("Settings.__init__(): data=%s", data)
                super().__init__(**data)
            except ValidationError as e:
                raise FlowSettingsError(
                    validation_errors(e.errors()), e.model  # type: ignore
                ) from None

    @property
    def succeeded(self) -> bool:
        return self.results.success

    def __init_subclass__(cls) -> None:
        cls_name = cls.__name__
        mod_name = cls.__module__
        log.info("registering flow %s from %s", cls_name, mod_name)
        cls.name = camelcase_to_snakecase(cls_name)
        if not inspect.isabstract(cls):
            registered_flows[cls.name] = (mod_name, cls)

    def init(self) -> None:
        """Flow custom initialization stage. At this point, more properties have been set than during __init__
        This is usually the most appropriate place for initialization task.
        Any dependent flows should be registered here by using add_dependency
        """

    def add_dependency(
        self, dep_flow_class: Type["Flow"], dep_settings: Settings
    ) -> None:
        self.dependencies.append((dep_flow_class, dep_settings))

    @classmethod
    def _create_jinja_env(
        cls, extra_modules: Optional[List[str]] = None
    ) -> jinja2.Environment:
        if extra_modules is None:
            extra_modules = []
        loaderChoices = []
        mod_paths = []
        modules = unique(
            extra_modules + [cls.__module__] + [clz.__module__ for clz in cls.__bases__]
        )
        for mpx in modules:
            for mp in [mpx, mpx.rsplit(".", 1)[0]]:
                if mp not in mod_paths:
                    mod_paths.append(mp)
        for mp in mod_paths:
            try:  # TODO better/cleaner way
                loaderChoices.append(PackageLoader(mp))
            except ValueError:
                pass
        return jinja2.Environment(
            loader=ChoiceLoader(loaderChoices),
            autoescape=False,
            undefined=StrictUndefined,
        )

    def __init__(self, settings: Settings, design: Design, run_path: Path):
        self.run_path = run_path
        self.settings = settings
        # assert isinstance(self.settings, self.Settings)
        self.design: Design = design

        self.init_time: Optional[float] = None
        self.timestamp: Optional[str] = None
        self.flow_hash: Optional[str] = None
        self.design_hash: Optional[str] = None

        # a map of "precious" files generated by the flow
        # these files can be used in a subsequent flow depending on this flow
        # everything else in run_path can be deleted by the flow-runner or overwritten in subsequent operations
        # artifacts can be backed up if the same run_path is used after this flow is complete
        self.artifacts = Box()
        self.reports: DictStrPath = {}
        # TODO deprecate and use self.reports
        self.reports_dir = run_path / self.settings.reports_dir
        self.reports_dir.mkdir(exist_ok=True)
        self.results: Box = Box(
            success=False,
            # "Time of the execution of run() in fractional seconds.
            # Initialized with None and set only after execution has finished."
            runtime=None,
            artifacts={},
        )
        self.jinja_env = self._create_jinja_env(extra_modules=[self.__module__])
        self.add_template_test("match", regex_match)
        self.dependencies: List[Tuple[Type[Flow], Flow.Settings]] = []
        self.completed_dependencies: List[Flow] = []

    def pop_dependency(self, typ: Type[T]) -> Flow:
        assert inspect.isclass(typ) and issubclass(typ, Flow)
        for i in range(len(self.completed_dependencies) - 1, -1, -1):
            if isinstance(self.completed_dependencies[i], typ):
                dep = self.completed_dependencies.pop(i)
                assert isinstance(dep, typ)
                return dep
        raise ValueError(f"No {typ.__name__} found in completed_dependencies")

    @abstractmethod
    def run(self) -> None:
        """return False on failure"""

    def parse_reports(self) -> bool:
        log.info("No parse_reports action for %s", self.name)
        return True

    def copy_from_template(self, resource_name, **kwargs) -> os.PathLike:
        template = self.jinja_env.get_template(resource_name)
        script_path: Path = self.run_path / resource_name
        log.debug("generating %s from template.", str(script_path.resolve()))
        rendered_content = template.render(
            settings=self.settings,
            design=self.design,
            artifacts=self.artifacts,
            **kwargs,
        )
        with open(script_path, "w") as f:
            f.write(rendered_content)
        return script_path.relative_to(self.run_path)  # resource_name

    def add_template_filter(self, filter_name: str, func) -> None:
        self.jinja_env.filters[filter_name] = func

    def add_template_test(self, filter_name: str, func) -> None:
        self.jinja_env.tests[filter_name] = func

    def conv_to_relative_path(self, src):
        path = Path(src).resolve(strict=True)
        return os.path.relpath(path, self.run_path)

    def parse_report_regex(
        self,
        reportfile_path: Union[str, os.PathLike],
        re_pattern: Union[str, List[str]],
        *other_re_patterns: Union[str, List[str]],
        dotall: bool = True,
        required: bool = False,
        sequential: bool = False,
    ) -> bool:
        if not isinstance(reportfile_path, Path):
            reportfile_path = Path(reportfile_path)
        # TODO fix debug and verbosity levels!
        if not reportfile_path.exists():
            log.warning(
                "File %s does not exist! Most probably the flow run had failed.\n Please check log files in %s",
                reportfile_path,
                self.run_path,
            )
            return False
        with open(reportfile_path) as rpt_file:
            content = rpt_file.read()

            flags = re.MULTILINE | re.IGNORECASE
            if dotall:
                flags |= re.DOTALL

            def match_pattern(pat: str, content: str) -> Tuple[bool, str]:
                match = re.search(pat, content, flags)
                if match is None:
                    return False, content
                match_dict = match.groupdict()
                for k, v in match_dict.items():
                    self.results[k] = try_convert(v)
                    log.debug("%s: %s", k, self.results.get(k))
                if sequential:
                    content = content[match.span(0)[1] :]
                    log.debug("len(content)=%d", len(content))
                return True, content

            for pat in [re_pattern, *other_re_patterns]:
                matched = False
                if isinstance(pat, list):
                    log.debug("Matching any of: %s", pat)
                    for subpat in pat:
                        matched, content = match_pattern(subpat, content)
                else:
                    log.debug("Matching: %s", pat)
                    matched, content = match_pattern(pat, content)

                if not matched and required:
                    log.critical(
                        "Error parsing report file: %s\n Pattern not matched: %s\n",
                        rpt_file.name,
                        pat,
                    )
                    return False
        return True


class SimFlow(Flow, metaclass=ABCMeta):
    """superclass of all simulation flows"""

    cocotb_sim_name: Optional[str] = None

    class Settings(Flow.Settings):
        vcd: Optional[str] = None
        stop_time: Union[None, str, int, float] = None
        cocotb: CocotbSettings = (
            CocotbSettings()
        )  # pyright: reportGeneralTypeIssues=none
        optimization_flags: List[str] = Field([], description="Optimization flags")

        @validator("vcd", pre=True)
        def validate_vcd(cls, vcd):  # pylint: disable=no-self-argument
            if vcd is not None:
                if isinstance(vcd, bool) or not vcd:
                    vcd = "dump.vcd" if vcd else None
                else:
                    assert isinstance(vcd, str)
                    if vcd[1:].count(".") == 0:  # if it doesn't have an extension
                        vcd += ".vcd"
            return vcd

    def __init__(self, settings: Settings, design: Design, run_path: Path):
        super().__init__(settings, design, run_path)
        assert isinstance(self.settings, self.Settings)
        self.cocotb: Optional[Cocotb] = (
            Cocotb(
                **self.settings.cocotb.dict(),
                sim_name=self.cocotb_sim_name,
            )
            if self.cocotb_sim_name
            else None
        )


class FPGA(XedaBaseModel):
    """FPGA target device"""

    # definition order: part > device > vendor > {family, speed, package, etc}
    part: Optional[str] = Field(None, description="full device part identifier")
    device: Optional[str]
    vendor: Optional[str]
    family: Optional[str]
    generation: Optional[str]
    type: Optional[str]
    speed: Optional[str] = Field(None, description="speed-grade")
    package: Optional[str]
    capacity: Optional[str]
    pins: Optional[int]
    grade: Optional[str]

    def __init__(self, *args: str, **data: Any) -> None:
        if args:
            if len(args) != 1 or not args[0]:
                raise ValueError("Only a single 'part' non-keyword argument is valid.")
            a = args[0]
            if isinstance(a, str):
                if "part" in data:
                    raise ValueError("'part' field already given in keyword arguments.")
                data["part"] = a
            elif isinstance(a, dict):
                if data:
                    raise ValueError("Both dictionary and keyword arguments preset")
                data = a
            else:
                raise ValueError(f"Argument of type {type(a)} is not supported!")
        log.debug("fpga init! data=%s", data)
        super().__init__(**data)

    # this is called before all field validators!
    @root_validator(pre=True)
    def fpga_root_validator(cls, values):  # pylint: disable=no-self-argument
        if not values:
            return values
        # Intel: https://www.intel.com/content/dam/www/central-libraries/us/en/documents/product-catalog.pdf
        # Lattice: https://www.latticesemi.com/Support/PartNumberReferenceGuide
        # Xilinx: https://www.xilinx.com/support/documents/selection-guides/7-series-product-selection-guide.pdf
        #         https://docs.xilinx.com/v/u/en-US/ds890-ultrascale-overview
        #         https://www.xilinx.com/support/documents/selection-guides/ultrascale-fpga-product-selection-guide.pdf
        #         https://www.xilinx.com/support/documents/selection-guides/ultrascale-plus-fpga-product-selection-guide.pdf
        part = values.get("part")

        def set_if_not_exist(attr: str, v: Any) -> None:
            if attr not in values:
                values[attr] = v

        def set_xc_family(s: str):
            d = dict(s="spartan", a="artix", k="kintex", v="virtex", z="zynq")
            s = s.lower()
            if s in d:
                set_if_not_exist("family", d[s])

        if part:
            part = part.strip()
            values["part"] = part
            # speed: 6 = slowest, 8 = fastest
            match_ecp5 = re.match(
                r"^LFE5(U|UM|UM5G)-(\d+)F-(?P<sp>\d)(?P<pkg>[A-Z]+)(?P<pin>\d+)(?P<gr>[A-Z]?)$",
                part,
                flags=re.IGNORECASE,
            )
            if match_ecp5:
                set_if_not_exist("vendor", "lattice")
                set_if_not_exist("family", "ecp5")
                set_if_not_exist("type", match_ecp5.group(1).lower())
                set_if_not_exist("device", "LFE5" + match_ecp5.group(1).upper())
                set_if_not_exist("capacity", match_ecp5.group(2) + "k")
                set_if_not_exist("speed", match_ecp5.group("sp"))
                set_if_not_exist("package", match_ecp5.group("pkg"))
                set_if_not_exist("pins", int(match_ecp5.group("pin")))
                set_if_not_exist("grade", match_ecp5.group("gr"))
                return values
            # Commercial Xilinx # Generation # Family # Logic Cells in 1K units # Speed Grade (-1 slowest, L: low-power) # Package Type
            match_xc7 = re.match(
                r"^(XC)(?P<g>\d)(?P<f>[A-Z])(?P<lc>\d+)-(?P<s>-L?\d)(?P<pkg>[A-Z][A-Z][A-Z]+)(?P<pin>\d\d+)(?P<gr>[A-Z]?)$",
                part,
                flags=re.IGNORECASE,
            )
            if match_xc7:
                set_if_not_exist("vendor", "xilinx")
                set_if_not_exist("generation", match_xc7.group("g"))
                set_xc_family(match_xc7.group("f"))
                lc = match_xc7.group("lc")
                set_if_not_exist(
                    "device",
                    match_xc7.group(1)
                    + match_xc7.group("g")
                    + match_xc7.group("f")
                    + lc,
                )
                set_if_not_exist("capacity", lc + "K")
                set_if_not_exist("package", int(match_xc7.group("pkg")))
                set_if_not_exist("pins", int(match_xc7.group("pins")))
                set_if_not_exist("grade", match_xc7.group("gr"))
                return values
            match_us = re.match(
                r"^(XC)(?P<f>[A-Z])(?P<g>[A-Z]+)(?P<lc>\d+)-(?P<s>-L?\d)(?P<pkg>[A-Z][A-Z][A-Z]+)(?P<pin>\d\d+)(?P<gr>[A-Z]?)$",
                part,
                flags=re.IGNORECASE,
            )
            if match_us:
                set_if_not_exist("vendor", "xilinx")
                set_if_not_exist("generation", match_us.group("g"))
                set_xc_family(match_us.group("f"))
                lc = match_us.group("lc")
                set_if_not_exist(
                    "device",
                    match_us.group(1) + match_us.group("g") + match_us.group("f") + lc,
                )
                set_if_not_exist("capacity", lc + "K")
                set_if_not_exist("package", int(match_us.group("pkg")))
                set_if_not_exist("pins", int(match_us.group("pins")))
                set_if_not_exist("grade", match_us.group("gr"))
                return values
            # UltraSCALE+
            # capacity is index to table, roughly x100K LCs
            match_usp = re.match(
                r"^(XC)(?P<f>[A-Z])U(?P<lc>\d+)P-(?P<s>-L?\d)(?P<pkg>[A-Z][A-Z][A-Z]+)(?P<pin>\d\d+)(?P<gr>[A-Z]?)$",
                part,
                flags=re.IGNORECASE,
            )
            if match_usp:
                set_if_not_exist("vendor", "xilinx")
                set_if_not_exist("generation", "usp")
                set_xc_family(match_usp.group("f"))
                lc = match_usp.group("lc")
                set_if_not_exist("capacity", lc)
                set_if_not_exist(
                    "device", match_usp.group(1) + match_usp.group("f") + "U" + lc + "P"
                )
                set_if_not_exist("package", int(match_usp.group("pkg")))
                set_if_not_exist("pins", int(match_usp.group("pins")))
                set_if_not_exist("grade", match_usp.group("gr"))
                return values
        elif not values.get("device") and not values.get("vendor"):
            raise ValueError(
                "Missing enough information about the FPGA device. Please set the 'part' number and/or device, vendor, family, etc."
            )
        return values


class TargetTechnology(XedaBaseModel):
    fpga: Optional[FPGA] = None
    liberty: Optional[str] = None
    gates: Optional[str] = None
    lut: Optional[str] = None


class PhysicalClock(XedaBaseModel):
    name: Optional[str] = None
    period: float = Field(
        description="Clock Period (in nanoseconds). Either (and only one of) 'period' OR 'freq' have to be specified."
    )
    freq: float = Field(
        description="Clock frequency (in MegaHertz). Either (and only one of) 'period' OR 'freq' have to be specified."
    )
    rise: float = Field(0.0, description="Rise time (nanoseconds)")
    fall: float = Field(0.0, description="Rall time (nanoseconds)")
    uncertainty: Optional[float] = Field(None, description="Clock uncertainty")
    skew: Optional[float] = Field(None, description="skew")
    port: Optional[str] = Field(None, description="associated design port")

    @validator("fall", always=True)
    def fall_validate(cls, value, values):  # pylint: disable=no-self-argument
        if not value:
            value = round(values.get("period", 0.0) / 2.0, 3)
        return value

    @property
    def duty_cycle(self) -> float:
        return (self.fall - self.rise) / self.period

    @property
    def freq_mhz(self) -> float:
        return 1000.0 / self.period

    @root_validator(pre=True, skip_on_failure=True)
    @classmethod
    def root_validate_phys_clock(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        log.debug("%s root_validator values=%s", cls.__name__, values)
        freq = values.get("freq")
        if "period" in values:
            period = float(values["period"])
            if freq and abs(float(freq) * period - 1000.0) > 0.001:
                raise ValueError(
                    f"Both freq and period cannot be specified at the same time period={values.get('period')} freq={values.get('freq')}"
                )
            values["freq"] = 1000.0 / values["period"]
        else:
            if freq:
                values["period"] = 1000.0 / float(freq)
            else:
                raise ValueError("Neither freq or period were specified")
        if not values.get("name"):
            values["name"] = "main_clock"
        return values


class SynthFlow(Flow, metaclass=ABCMeta):
    """Superclass of all synthesis flows"""

    class Settings(Flow.Settings):
        """base Synthesis flow settings"""

        clock_period: Optional[float] = Field(
            None, description="target clock period in nanoseconds"
        )
        clocks: Dict[str, PhysicalClock] = {}
        fpga: Optional[FPGA] = None
        tech: Optional[TargetTechnology] = None
        blacklisted_resources: Optional[List[str]]

        @validator("clocks", always=True)
        @classmethod
        def clocks_validate(cls, value, values):  # pylint: disable=no-self-argument
            clock_period = values.get("clock_period")
            if not value and clock_period:
                value = {
                    "main_clock": PhysicalClock(name="main_clock", period=clock_period)  # type: ignore
                }
            return value

        @root_validator()
        @classmethod
        def synthflow_settings_root_validator(cls, values):
            clocks = values.get("clocks")
            if clocks and "main_clock" in clocks and not values.get("clock_period"):
                main_clock = clocks["main_clock"]
                assert isinstance(main_clock, PhysicalClock)
                values["clock_period"] = main_clock.period
            return values

    def __init__(self, flow_settings: Settings, design: Design, run_path: Path):
        for clock_name, physical_clock in flow_settings.clocks.items():
            if not physical_clock.port:
                if clock_name not in design.rtl.clocks:
                    raise FlowSettingsError(
                        [
                            (
                                None,
                                f"Physical clock {clock_name} has no corresponding clock port in design. Existing clocks: {', '.join(c for c in design.rtl.clocks)}",
                                None,
                            )
                        ],
                        self.Settings,
                    )
                physical_clock.port = design.rtl.clocks[clock_name].port
                flow_settings.clocks[clock_name] = physical_clock
        for clock_name, clock in design.rtl.clocks.items():
            if clock_name not in flow_settings.clocks:
                raise FlowSettingsError(
                    [
                        (
                            None,
                            f"No clock period or frequency was specified for clock: '{clock_name}' (clock port: '{clock.port})'",
                            None,
                        )
                    ],
                    self.Settings,
                )
        super().__init__(flow_settings, design, run_path)


class FpgaSynthFlow(SynthFlow, metaclass=ABCMeta):
    """Superclass of all FPGA synthesis flows"""

    class Settings(SynthFlow.Settings):
        """base FPGA Synthesis flow settings"""

        fpga: FPGA


class AsicSynthFlow(SynthFlow, metaclass=ABCMeta):
    """Superclass of all ASIC synthesis flows"""

    class Settings(SynthFlow.Settings):
        """base ASIC Synthesis flow settings"""


class DseFlow(Flow, metaclass=ABCMeta):
    """Superclass of all design-space exploration flows"""


class FlowException(Exception):
    """Super-class of all flow exceptions"""


class FlowDependencyFailure(FlowException):
    """Error during execution of a dependency flow"""


class FlowSettingsError(FlowException):
    """Validation of settings failed"""

    def __init__(
        self,
        errors: List[
            Tuple[Optional[str], str, Optional[str]]
        ],  # (location, message, type/context)
        model,
        *args: Any,
    ) -> None:
        super().__init__(*args)
        self.errors = errors
        self.model = model

    def __str__(self) -> str:
        return "{}: {} error{} validating {}:\n{}".format(
            self.__class__.__qualname__,
            len(self.errors),
            "s" if len(self.errors) > 1 else "",
            self.model.__qualname__,
            "\n".join(
                "{}{}\n".format(f"{loc}:\n   " if loc else "", msg)
                for loc, msg, ctx in self.errors
            ),
        )


class FlowFatalError(FlowException):
    """Other fatal errors"""



import logging
from abc import ABCMeta
from functools import reduce
from html import unescape
from pathlib import Path
from typing import Any, Dict, Optional
from xml.etree import ElementTree


from ...dataclass import Field
from ...design import Design
from ...tool import Docker, Tool
from ..flow import Flow

log = logging.getLogger(__name__)


def vivado_generics(kvdict, sim=False):
    def supported_vivado_generic(v, sim):
        if sim:
            return True
        if isinstance(v, int):
            return True
        if isinstance(v, bool):
            return True
        v = str(v)
        return v.isnumeric() or (v.strip().lower() in {"true", "false"})

    def vivado_gen_convert(k, x, sim):
        if sim:
            return x
        xl = str(x).strip().lower()
        if xl == "false":
            return "1\\'b0"
        if xl == "true":
            return "1\\'b1"
        return x

    return " ".join(
        [
            f"-generic{'_top' if sim else ''} {k}={vivado_gen_convert(k, v, sim)}"
            for k, v in kvdict.items()
            if supported_vivado_generic(v, sim)
        ]
    )


class Vivado(Flow, metaclass=ABCMeta):
    """Xilinx (AMD) Vivado FPGA synthesis and simulation flows"""

    class Settings(Flow.Settings):
        tcl_shell: bool = Field(
            False,
            description="Drop to interactive TCL shell after Vivado finishes running a flow script",
        )

    def __init__(self, settings: Settings, design: Design, run_path: Path):
        super().__init__(settings, design, run_path)
        default_args = [
            "-nojournal",
            "-mode",
            "tcl" if settings.tcl_shell else "batch",
        ]
        if not self.settings.debug:
            default_args.append("-notrace")
        self.vivado = Tool("vivado", default_args=default_args, docker=Docker(image="pwang7/vivado_ubuntu", tag="standard-2021.2", enabled=self.settings.dockerized))
        self.add_template_filter("vivado_generics", vivado_generics)

    @staticmethod
    def parse_xml_report(report_xml) -> Optional[Dict[str, Any]]:
        try:
            tree = ElementTree.parse(report_xml)
        except ElementTree.ParseError as e:
            log.critical("Parsing %s failed: %s", report_xml, e.msg)
            return None
        data = {}
        for section in tree.findall("./section"):
            section_title = section.get("title", "<section>")
            for table in section.findall("./table"):
                table_data = {}
                header = [
                    unescape(col.attrib["contents"]).strip()
                    for col in table.findall("./tablerow/tableheader")
                ]
                for tablerow in table.findall("./tablerow"):
                    cells = [
                        unescape(cell.attrib["contents"]).strip()
                        for cell in tablerow.findall("./tablecell")
                    ]
                    if cells:
                        # choose 0th element as "index data" (distinct key)
                        cell_data = {h: c for h, c in zip(header[1:], cells[1:]) if c}
                        if cell_data:
                            table_data[cells[0]] = cell_data
                if table_data:
                    table_title = table.get("title")
                    title = (
                        section_title + ":" + table_title
                        if table_title
                        else section_title
                    )
                    data[title] = table_data
        return data

    @staticmethod
    def get_from_path(dct: dict, path):
        if isinstance(path, str):
            path = path.split(".")
        return reduce(dict.__getitem__, path, dct)
